"""
JAX 版 RWKV7 Triton Kernel 封装
利用 jax-triton 直接调用 triton_kernel.py 中的算子，并支持 SPMD 切分
"""

from __future__ import annotations
import jax
import jax.numpy as jnp
import jax_triton as jt
import jax.tree_util as jtu
from typing import Optional, Tuple, Union

# 引入 Triton 核心算子
from .triton_kernel import (
    rwkv7_fwd_kernel,
    rwkv7_bwd_kernel,
    rwkv7_fwd_kernel_with_mask,
    rwkv7_bwd_kernel_with_mask,
)

# 引入自定义分区器 (适配 JAX 新版 Shardy 引擎)
from jax.experimental.custom_partitioning import custom_partitioning

CHUNK_LEN = 16

# =========================================================================
# SPMD 切分规则 (Einsum 风格)
# b=Batch, n=Head, t=Time, h=HeadDim1, m=HeadDim2, c=Chunk
# =========================================================================
# 无 Mask 规则
FWD_RULE = "b n t h, b n t h, b n t h, b n t h, b n t h, b n t h, b n h m -> b n t h, b n t h, b n c h m"
BWD_RULE = "b n t h, b n t h, b n t h, b n t h, b n t h, b n t h, b n t h, b n t h, b n c h m, b n h m -> b n t h, b n t h, b n t h, b n t h, b n t h, b n t h, b n h m"

# 带 Mask 规则
FWD_MASK_RULE = "b n t h, b n t h, b n t h, b n t h, b n t h, b n t h, b n h m, b t -> b n t h, b n t h, b n c h m"
BWD_MASK_RULE = "b n t h, b n t h, b n t h, b n t h, b n t h, b n t h, b t, b n t h, b n t h, b n c h m, b n h m -> b n t h, b n t h, b n t h, b n t h, b n t h, b n t h, b n h m"


def _fwd_infer_sharding(arg_shapes, arg_shardings):
    qs = arg_shardings[0]  # 与输入对齐
    return (qs, qs, qs)


def _bwd_infer_sharding(arg_shapes, arg_shardings):
    qs = arg_shardings[0]
    return (qs, qs, qs, qs, qs, qs, qs)


def _create_partition(impl_fn):
    def partition(mesh, arg_shapes, result_shape):
        def lower_fn(*args):
            return impl_fn(*args)

        result_shardings = jtu.tree_map(lambda x: x.sharding, result_shape)
        arg_shardings = jtu.tree_map(lambda x: x.sharding, arg_shapes)
        return mesh, lower_fn, result_shardings, arg_shardings

    return partition


def _transpose_head(x: jnp.ndarray, head_first: bool) -> jnp.ndarray:
    x = jnp.asarray(x, dtype=jnp.bfloat16)
    if not head_first:
        # [B, T, N, H] -> [B, N, T, H]
        return jnp.transpose(x, (0, 2, 1, 3))
    return x


# =========================================================================
# 无 Mask 的 JAX-Triton Launcher
# =========================================================================
def _wkv7_fwd_triton_call(r, w, k, v, a, b, h0):
    B, N, T, H = r.shape
    dtype = r.dtype
    chunk_num = T // CHUNK_LEN

    out_shapes = [
        jax.ShapeDtypeStruct((B, N, T, H), dtype),  # OUT
        jax.ShapeDtypeStruct((B, N, T, H), jnp.float32),  # SA_OUT
        jax.ShapeDtypeStruct((B, N, chunk_num, H, H), jnp.float32),  # STATE_CHKP
    ]

    grid = lambda meta: ((B + meta["MINI_BSZ"] - 1) // meta["MINI_BSZ"], N)

    out, sa_out, state_chkp = jt.triton_call(
        r,
        w,
        k,
        v,
        a,
        b,
        h0,
        B,
        N,
        T,
        kernel=rwkv7_fwd_kernel,
        out_shape=out_shapes,
        grid=grid,
        H_SIZE=H,
        CHUNK_LEN=CHUNK_LEN,
    )
    return out, sa_out, state_chkp


@custom_partitioning
def _wkv7_fwd_spmd(r, w, k, v, a, b, h0):
    return _wkv7_fwd_triton_call(r, w, k, v, a, b, h0)


_wkv7_fwd_spmd.def_partition(
    infer_sharding_from_operands=_fwd_infer_sharding,
    sharding_rule=FWD_RULE,
    partition=_create_partition(_wkv7_fwd_triton_call),
)


def _wkv7_bwd_triton_call(r, w, k, v, a, b, dy, sa, state_chkp, dht):
    B, N, T, H = r.shape
    dtype = r.dtype

    out_shapes = [
        jax.ShapeDtypeStruct((B, N, T, H), dtype),  # DR
        jax.ShapeDtypeStruct((B, N, T, H), dtype),  # DW
        jax.ShapeDtypeStruct((B, N, T, H), dtype),  # DK
        jax.ShapeDtypeStruct((B, N, T, H), dtype),  # DV
        jax.ShapeDtypeStruct((B, N, T, H), dtype),  # DA
        jax.ShapeDtypeStruct((B, N, T, H), dtype),  # DB
        jax.ShapeDtypeStruct((B, N, H, H), jnp.float32),  # DH0
    ]

    grid = lambda meta: ((B + meta["MINI_BSZ"] - 1) // meta["MINI_BSZ"], N)

    dr, dw, dk, dv, da, db, dh0 = jt.triton_call(
        r,
        w,
        k,
        v,
        a,
        b,
        sa,
        state_chkp,
        B,
        N,
        T,
        dy,
        dht,
        kernel=rwkv7_bwd_kernel,
        out_shape=out_shapes,
        grid=grid,
        H_SIZE=H,
        CHUNK_LEN=CHUNK_LEN,
    )
    return dr, dw, dk, dv, da, db, dh0


@custom_partitioning
def _wkv7_bwd_spmd(r, w, k, v, a, b, dy, sa, state_chkp, dht):
    return _wkv7_bwd_triton_call(r, w, k, v, a, b, dy, sa, state_chkp, dht)


_wkv7_bwd_spmd.def_partition(
    infer_sharding_from_operands=_bwd_infer_sharding,
    sharding_rule=BWD_RULE,
    partition=_create_partition(_wkv7_bwd_triton_call),
)


@jax.custom_vjp
def rwkv7_kernel_triton(r, w, k, v, a, b, h0):
    out, sa_out, state_chkp = _wkv7_fwd_spmd(r, w, k, v, a, b, h0)
    final_state = state_chkp[:, :, -1, :, :]
    return out, final_state


def _fwd(r, w, k, v, a, b, h0):
    out, sa_out, state_chkp = _wkv7_fwd_spmd(r, w, k, v, a, b, h0)
    final_state = state_chkp[:, :, -1, :, :]
    return (out, final_state), (r, w, k, v, a, b, sa_out, state_chkp)


def _bwd(res, grads):
    r, w, k, v, a, b, sa_out, state_chkp = res
    dy, dht = grads
    dy = jnp.asarray(dy, jnp.bfloat16)
    if dht is None:
        B, N, T, H = r.shape
        dht = jnp.zeros((B, N, H, H), dtype=jnp.float32)
    else:
        dht = jnp.asarray(dht, jnp.float32)

    dr, dw, dk, dv, da, db, dh0 = _wkv7_bwd_spmd(
        r, w, k, v, a, b, dy, sa_out, state_chkp, dht
    )
    return dr, dw, dk, dv, da, db, dh0


rwkv7_kernel_triton.defvjp(_fwd, _bwd)


# =========================================================================
# 带 Mask 的 JAX-Triton Launcher
# =========================================================================
def _wkv7_fwd_with_mask_triton_call(r, w, k, v, a, b, h0, mask):
    B, N, T, H = r.shape
    dtype = r.dtype
    chunk_num = T // CHUNK_LEN

    out_shapes = [
        jax.ShapeDtypeStruct((B, N, T, H), dtype),
        jax.ShapeDtypeStruct((B, N, T, H), jnp.float32),
        jax.ShapeDtypeStruct((B, N, chunk_num, H, H), jnp.float32),
    ]

    grid = lambda meta: ((B + meta["MINI_BSZ"] - 1) // meta["MINI_BSZ"], N)

    out, sa_out, state_chkp = jt.triton_call(
        r,
        w,
        k,
        v,
        a,
        b,
        mask,
        h0,
        B,
        N,
        T,
        kernel=rwkv7_fwd_kernel_with_mask,
        out_shape=out_shapes,
        grid=grid,
        H_SIZE=H,
        CHUNK_LEN=CHUNK_LEN,
    )
    return out, sa_out, state_chkp


@custom_partitioning
def _wkv7_fwd_with_mask_spmd(r, w, k, v, a, b, h0, mask):
    return _wkv7_fwd_with_mask_triton_call(r, w, k, v, a, b, h0, mask)


_wkv7_fwd_with_mask_spmd.def_partition(
    infer_sharding_from_operands=_fwd_infer_sharding,
    sharding_rule=FWD_MASK_RULE,
    partition=_create_partition(_wkv7_fwd_with_mask_triton_call),
)


def _wkv7_bwd_with_mask_triton_call(r, w, k, v, a, b, mask, dy, sa, state_chkp, dht):
    B, N, T, H = r.shape
    dtype = r.dtype

    out_shapes = [
        jax.ShapeDtypeStruct((B, N, T, H), dtype),  # DR
        jax.ShapeDtypeStruct((B, N, T, H), dtype),  # DW
        jax.ShapeDtypeStruct((B, N, T, H), dtype),  # DK
        jax.ShapeDtypeStruct((B, N, T, H), dtype),  # DV
        jax.ShapeDtypeStruct((B, N, T, H), dtype),  # DA
        jax.ShapeDtypeStruct((B, N, T, H), dtype),  # DB
        jax.ShapeDtypeStruct((B, N, H, H), jnp.float32),  # DH0
    ]

    grid = lambda meta: ((B + meta["MINI_BSZ"] - 1) // meta["MINI_BSZ"], N)

    dr, dw, dk, dv, da, db, dh0 = jt.triton_call(
        r,
        w,
        k,
        v,
        a,
        b,
        mask,
        sa,
        state_chkp,
        B,
        N,
        T,
        dy,
        dht,
        kernel=rwkv7_bwd_kernel_with_mask,
        out_shape=out_shapes,
        grid=grid,
        H_SIZE=H,
        CHUNK_LEN=CHUNK_LEN,
    )
    return dr, dw, dk, dv, da, db, dh0


@custom_partitioning
def _wkv7_bwd_with_mask_spmd(r, w, k, v, a, b, mask, dy, sa, state_chkp, dht):
    return _wkv7_bwd_with_mask_triton_call(
        r, w, k, v, a, b, mask, dy, sa, state_chkp, dht
    )


_wkv7_bwd_with_mask_spmd.def_partition(
    infer_sharding_from_operands=_bwd_infer_sharding,
    sharding_rule=BWD_MASK_RULE,
    partition=_create_partition(_wkv7_bwd_with_mask_triton_call),
)


@jax.custom_vjp
def rwkv7_kernel_with_mask_triton(r, w, k, v, a, b, h0, mask):
    out, sa_out, state_chkp = _wkv7_fwd_with_mask_spmd(r, w, k, v, a, b, h0, mask)
    final_state = state_chkp[:, :, -1, :, :]
    return out, final_state


def _fwd_with_mask(r, w, k, v, a, b, h0, mask):
    out, sa_out, state_chkp = _wkv7_fwd_with_mask_spmd(r, w, k, v, a, b, h0, mask)
    final_state = state_chkp[:, :, -1, :, :]
    return (out, final_state), (r, w, k, v, a, b, mask, sa_out, state_chkp)


def _bwd_with_mask(res, grads):
    r, w, k, v, a, b, mask, sa_out, state_chkp = res
    dy, dht = grads
    dy = jnp.asarray(dy, jnp.bfloat16)
    if dht is None:
        B, N, T, H = r.shape
        dht = jnp.zeros((B, N, H, H), dtype=jnp.float32)
    else:
        dht = jnp.asarray(dht, jnp.float32)

    dr, dw, dk, dv, da, db, dh0 = _wkv7_bwd_with_mask_spmd(
        r, w, k, v, a, b, mask, dy, sa_out, state_chkp, dht
    )
    return dr, dw, dk, dv, da, db, dh0, None


rwkv7_kernel_with_mask_triton.defvjp(_fwd_with_mask, _bwd_with_mask)


# =========================================================================
# 对外 API 暴露 (与 JAX/Torch CUDA 版本兼容对齐)
# =========================================================================
def generalized_delta_rule(
    r: jnp.ndarray,
    w: jnp.ndarray,
    k: jnp.ndarray,
    v: jnp.ndarray,
    a: jnp.ndarray,
    b: jnp.ndarray,
    initial_state: Optional[jnp.ndarray] = None,
    output_final_state: bool = True,
    head_first: bool = False,
    mask: Optional[jnp.ndarray] = None,
) -> Union[jnp.ndarray, Tuple[jnp.ndarray, jnp.ndarray]]:
    # 统一转换到 Head-First [B, N, T, H]
    r = _transpose_head(r, head_first)
    w = _transpose_head(w, head_first)
    k = _transpose_head(k, head_first)
    v = _transpose_head(v, head_first)
    a = _transpose_head(a, head_first)
    b = _transpose_head(b, head_first)

    B, N, T, H = r.shape
    if T % CHUNK_LEN != 0:
        raise ValueError(
            f"Triton kernel requires sequence length T={T} to be divisible by {CHUNK_LEN}"
        )

    # 准备初始状态
    if initial_state is None:
        h0 = jnp.zeros((B, N, H, H), dtype=jnp.float32)
    else:
        h0 = jnp.asarray(initial_state, dtype=jnp.float32)

    # 路由调用 Mask 还是 Non-Mask
    if mask is None:
        out, last_state = rwkv7_kernel_triton(r, w, k, v, a, b, h0)
    else:
        if mask.shape != (B, T) and mask.shape != (B, T, 1, 1):
            raise ValueError(
                f"Mask shape must be (B, T) or (B, T, 1, 1), got {mask.shape}"
            )
        mask = jnp.asarray(mask, dtype=jnp.float32).reshape(B, T)
        out, last_state = rwkv7_kernel_with_mask_triton(r, w, k, v, a, b, h0, mask)

    out = jnp.transpose(out, (0, 2, 1, 3))

    if output_final_state:
        return out, last_state
    return out



from keras import ops


class RWKVKernelOperator:
    def __init__(self, head_size, max_sequence_length):
        self.head_size = head_size
        self.max_sequence_length = max_sequence_length

    def __call__(
        self, r, k, v, w, u, with_state=False, init_state=None, state_map=None
    ):
        B, T, C = ops.shape(r)
        assert C % self.head_size == 0
        H = C // self.head_size
        w = ops.reshape(w, [B, T, H, self.head_size, 1])
        k = ops.reshape(k, [B, T, H, self.head_size, 1])

        v = ops.reshape(v, [B, T, H, 1, self.head_size])
        r = ops.reshape(r, [B, T, H, 1, self.head_size])
        u = ops.reshape(u, [1, H, self.head_size, 1])

        if init_state is not None:
            assert len(init_state.shape) in [
                3,
                4,
            ], "init_state的形状必须为(state_kinds,num_heads,head_size,head_size)"
            if len(init_state.shape) == 3:
                assert init_state.shape == (
                    H,
                    self.head_size,
                    self.head_size,
                ), "state_kinds的形状必须为(BatchSize,num_heads,head_size,head_size)"
                init_state = init_state[None, :]
            else:
                assert init_state.shape[1:] == (
                    H,
                    self.head_size,
                    self.head_size,
                ), "state_kinds的形状必须为(BatchSize,num_heads,head_size,head_size)"
                state_kinds = init_state.shape[0]
            if state_map is None:
                state_kinds = init_state.shape[0]
                if state_kinds == 1:
                    state_map = ops.zeros(shape=(B,), dtype="int32")
                elif state_kinds == B:
                    state_map = ops.convert_to_tensor(
                        [i for i in range(B)], dtype="int32"
                    )
                else:
                    raise ValueError(
                        "无法为您推断state_map的形状，请您手动指定state_map"
                    )

            else:
                if isinstance(state_map, list):
                    state_map = ops.convert_to_tensor(state_map, dtype="int32")
                state_map = ops.cast(state_map, "int32")
                assert (state_map >= 0).all() and (state_map < state_kinds).all(), (
                    f"请确保state_map的值域为[0, {state_kinds})"
                )
            s = ops.take(init_state, state_map, axis=0)

        else:
            assert state_map is None
            s = ops.zeros((B, H, self.head_size, self.head_size), dtype=u.dtype)

        w = ops.exp(-ops.exp(w))

        def cond(i, k, v, w, r, s, y):
            return i < T

        def body(i, k, v, w, r, s, y):
            k_t = ops.take(k, i, 1)
            v_t = ops.take(v, i, 1)
            kv_t = k_t @ v_t
            w_t = ops.take(w, i, 1)

            r_t = ops.take(r, i, 1)
            y_t = r_t @ (u * kv_t + s)
            y_t = ops.reshape(y_t, (B, 1, C))
            s = kv_t + w_t * s

            y = ops.slice_update(y, [0, i, 0], y_t)
            return i + 1, k, v, w, r, s, y

        y = ops.zeros([B, T, C], r.dtype)
        i, k, v, w, r, s, y = ops.while_loop(cond, body, (0, k, v, w, r, s, y), T)
        if with_state:
            return y, s
        return y, None



import os
import torch
from torch.utils.cpp_extension import load
from keras.src.backend.torch.core import cast
from keras.src.backend.torch.numpy import transpose


def transpose_head(x, head_first):
    if head_first:
        return transpose(x, (0, 2, 1, 3))
    else:
        return x


def get_torch_generalized_delta_rule(HEAD_SIZE=64):
    CHUNK_LEN = 16
    flags = [
        "-res-usage",
        f"-D_C_={HEAD_SIZE}",
        f"-D_CHUNK_LEN_={CHUNK_LEN}",
        "--use_fast_math",
        "-O3",
        "-Xptxas -O3",
        "--extra-device-vectorization",
    ]

    current_file_path = os.path.abspath(__file__)
    current_dir_path = os.path.dirname(current_file_path)

    load(
        name="wind_backstepping",
        sources=[
            os.path.join(current_dir_path, "wkv7_cuda.cu"),
            os.path.join(current_dir_path, "wkv7_op.cpp"),
        ],
        is_python_module=False,
        verbose=True,
        extra_cuda_cflags=flags,
    )

    # ============================================================
    # 原版无 Mask Autograd Function（内部使用）
    # ============================================================
    class WindBackstepping(torch.autograd.Function):
        @staticmethod
        def forward(ctx, w, q, k, v, a, b, h0):
            B, T, H, N = w.shape
            DTYPE = q.dtype
            q, k, v, a, b, w = [
                cast(x, "bfloat16").contiguous() for x in [q, k, v, a, b, w]
            ]

            if T % CHUNK_LEN != 0:
                raise ValueError("RWKV inputs sequence length must be divisible by 16")

            y = torch.empty_like(v)
            s = torch.empty(
                B, H, T // CHUNK_LEN, N, N, dtype=torch.float32, device=w.device
            )
            sa = torch.empty(B, T, H, N, dtype=torch.float32, device=w.device)

            # 注意原版接口：第5个参数是z(对应a)，第6个是a(对应b)
            torch.ops.wind_backstepping.forward(w, q, k, v, a, b, y, s, sa, h0)

            ctx.save_for_backward(w, q, k, v, a, b, s, sa)

            last_state = torch.empty_like(h0)
            last_state.copy_(transpose(s[:, :, -1], [0, 1, 3, 2]))
            return cast(y, DTYPE), last_state

        @staticmethod
        def backward(ctx, dy, dht):
            DTYPE = dy.dtype
            dy = cast(dy, torch.bfloat16).contiguous()
            dht = cast(dht, "float32").contiguous()
            w, q, k, v, a, b, s, sa = ctx.saved_tensors

            dh0 = torch.empty(dht.shape, dtype=dht.dtype, device=dht.device)
            dw, dq, dk, dv, da, db = [torch.empty_like(x) for x in [w, q, k, v, a, b]]

            # 原版接口：第5个输出是dz(对应da)，第6个是da(对应db)
            torch.ops.wind_backstepping.backward(
                w, q, k, v, a, b, dy, s, sa, dht, dh0, dw, dq, dk, dv, da, db
            )
            return (
                cast(dw, DTYPE),
                cast(dq, DTYPE),
                cast(dk, DTYPE),
                cast(dv, DTYPE),
                cast(da, DTYPE),
                cast(db, DTYPE),
                dh0,
            )

    # ============================================================
    # 带 Mask Autograd Function（内部使用）
    # ============================================================
    class WindBacksteppingWithMask(torch.autograd.Function):
        @staticmethod
        def forward(ctx, w, q, k, v, a, b, mask, h0):
            B, T, H, N = w.shape
            DTYPE = q.dtype

            q, k, v, a, b, w = [
                cast(x, "bfloat16").contiguous() for x in [q, k, v, a, b, w]
            ]
            mask = cast(mask, "bfloat16").contiguous()

            if T % CHUNK_LEN != 0:
                raise ValueError("RWKV inputs sequence length must be divisible by 16")

            y = torch.empty_like(v)
            s = torch.empty(
                B, H, T // CHUNK_LEN, N, N, dtype=torch.float32, device=w.device
            )
            sa = torch.empty(B, T, H, N, dtype=torch.float32, device=w.device)

            # Mask版本接口：参数直接对应 w,q,k,v,a,b,mask
            torch.ops.wind_backstepping.forward_with_mask(
                w, q, k, v, a, b, mask, y, s, sa, h0
            )

            ctx.save_for_backward(w, q, k, v, a, b, mask, s, sa)

            last_state = torch.empty_like(h0)
            last_state.copy_(transpose(s[:, :, -1], [0, 1, 3, 2]))
            return cast(y, DTYPE), last_state

        @staticmethod
        def backward(ctx, dy, dht):
            DTYPE = dy.dtype
            dy = cast(dy, torch.bfloat16).contiguous()
            dht = cast(dht, "float32").contiguous()

            w, q, k, v, a, b, mask, s, sa = ctx.saved_tensors

            dh0 = torch.empty(dht.shape, dtype=dht.dtype, device=dht.device)
            dw, dq, dk, dv, da, db = [torch.empty_like(x) for x in [w, q, k, v, a, b]]

            torch.ops.wind_backstepping.backward_with_mask(
                w, q, k, v, a, b, mask, dy, s, sa, dht, dh0, dw, dq, dk, dv, da, db
            )
            return (
                cast(dw, DTYPE),
                cast(dq, DTYPE),
                cast(dk, DTYPE),
                cast(dv, DTYPE),
                cast(da, DTYPE),
                cast(db, DTYPE),
                None,
                dh0,  # mask的梯度为None
            )

    # ============================================================
    # 统一对外接口：Training（根据mask自动选择）
    # ============================================================
    def generalized_delta_rule(
        r,
        w,
        k,
        v,
        a,
        b,
        initial_state=None,
        output_final_state: bool = True,
        head_first: bool = False,
        mask=None,
    ):
        """
        统一的RWKV7 Delta Rule前向/反向接口
        Args:
            mask: None时使用无mask版本，否则使用mask版本（自动转换为fp32）
        """
        # CPU回退
        if w.device.type != "cuda":
            from ..native_keras_op import generalized_delta_rule

            return generalized_delta_rule(
                r=r,
                w=w,
                k=k,
                v=v,
                a=a,
                b=b,
                mask=mask,
                initial_state=initial_state,
                output_final_state=output_final_state,
            )

        # 维度转置
        r = transpose_head(r, head_first)
        k = transpose_head(k, head_first)
        v = transpose_head(v, head_first)
        a = transpose_head(a, head_first)
        b = transpose_head(b, head_first)
        w = transpose_head(w, head_first)

        B, T, H, N = w.shape
        if initial_state is None:
            initial_state = torch.zeros(
                B, H, N, N, dtype=torch.float32, device=r.device
            )
        else:
            initial_state = cast(initial_state, "float32")

        # 统一处理：将Python参数名映射到Kernel参数
        # r->q, w->w, k->k, v->v, a->a/b(视版本而定), b->b/a(视版本而定)
        if mask is None:
            # 无mask版本：参数映射为 w,q,k,v,a(z),b
            out, state = WindBackstepping.apply(w, r, k, v, a, b, initial_state)
        else:
            # 带mask版本：需要确保mask是float32且在cuda上
            if not mask.is_cuda:
                mask = mask.to(torch.bfloat16)
            out, state = WindBacksteppingWithMask.apply(
                w, r, k, v, a, b, mask, initial_state
            )

        return (out, state) if output_final_state else out

    # ============================================================
    # 原版推理（内部）
    # ============================================================
    class Wkv7Inference(torch.autograd.Function):
        @staticmethod
        def forward(ctx, w, q, k, v, a, b, h0):
            B, T, H, N = w.shape
            DTYPE = q.dtype
            q, k, v, a, b, w = [
                cast(x, "bfloat16").contiguous() for x in [q, k, v, a, b, w]
            ]
            y = torch.empty_like(v)
            s = torch.empty(B, H, N, N, dtype=torch.float32, device=w.device)
            torch.ops.wind_backstepping.forward_inference(w, q, k, v, a, b, y, s, h0)
            return cast(y, DTYPE), s

        @staticmethod
        def backward(ctx, *args):
            raise NotImplementedError

    # ============================================================
    # 带Mask推理（内部）
    # ============================================================
    class Wkv7InferenceWithMask(torch.autograd.Function):
        @staticmethod
        def forward(ctx, w, q, k, v, a, b, mask, h0):
            B, T, H, N = w.shape
            DTYPE = q.dtype
            q, k, v, a, b, w = [
                cast(x, "bfloat16").contiguous() for x in [q, k, v, a, b, w]
            ]
            mask = cast(mask, "bfloat16").contiguous()
            y = torch.empty_like(v)
            s = torch.empty(B, H, N, N, dtype=torch.float32, device=w.device)
            torch.ops.wind_backstepping.forward_inference_with_mask(
                w, q, k, v, a, b, mask, y, s, h0
            )
            return cast(y, DTYPE), s

        @staticmethod
        def backward(ctx, *args):
            raise NotImplementedError

    def generalized_delta_rule_inference(
        r,
        w,
        k,
        v,
        a,
        b,
        initial_state=None,
        head_first: bool = False,
        output_final_state: bool = True,
        mask=None,
    ):
        if w.device.type != "cuda":
            raise NotImplementedError("Inference kernel only supports CUDA")

        r = transpose_head(r, head_first)
        k = transpose_head(k, head_first)
        v = transpose_head(v, head_first)
        a = transpose_head(a, head_first)
        b = transpose_head(b, head_first)
        w = transpose_head(w, head_first)

        B, T, H, N = w.shape
        if initial_state is None:
            initial_state = torch.zeros(
                B, H, N, N, dtype=torch.float32, device=r.device
            )
        else:
            initial_state = cast(initial_state, "float32")

        if mask is None:
            out, final_state = Wkv7Inference.apply(w, r, k, v, a, b, initial_state)
        else:
            if not mask.is_cuda:
                mask = mask.to(torch.bfloat16)
            out, final_state = Wkv7InferenceWithMask.apply(
                w, r, k, v, a, b, mask, initial_state
            )

        return (out, final_state) if output_final_state else out

    return [generalized_delta_rule, generalized_delta_rule_inference]
